"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _TreeCheckout_transaction;
Object.defineProperty(exports, "__esModule", { value: true });
exports.TreeCheckout = exports.createTreeCheckout = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const client_utils_1 = require("@fluid-internal/client-utils");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../codec/index.js");
const index_js_2 = require("../core/index.js");
const index_js_3 = require("../feature-libraries/index.js");
const index_js_4 = require("../shared-tree-core/index.js");
const index_js_5 = require("../util/index.js");
const sharedTreeChangeFamily_js_1 = require("./sharedTreeChangeFamily.js");
const schematizingTreeView_js_1 = require("./schematizingTreeView.js");
/**
 * Creates a {@link TreeCheckout}.
 * @param args - an object containing optional components that will be used to build the view.
 * Any components not provided will be created by default.
 * @remarks This does not create a {@link SharedTree}, but rather a view with the minimal state
 * and functionality required to implement {@link ITreeCheckout}.
 */
function createTreeCheckout(idCompressor, mintRevisionTag, revisionTagCodec, args) {
    const forest = args?.forest ?? (0, index_js_3.buildForest)();
    const schema = args?.schema ?? new index_js_2.TreeStoredSchemaRepository();
    const defaultCodecOptions = { jsonValidator: index_js_1.noopValidator };
    const defaultFieldBatchVersion = 1;
    const changeFamily = args?.changeFamily ??
        new sharedTreeChangeFamily_js_1.SharedTreeChangeFamily(revisionTagCodec, args?.fieldBatchCodec ??
            (0, index_js_3.makeFieldBatchCodec)(defaultCodecOptions, defaultFieldBatchVersion), { jsonValidator: index_js_1.noopValidator }, args?.chunkCompressionStrategy, idCompressor);
    const branch = args?.branch ??
        new index_js_4.SharedTreeBranch({
            change: changeFamily.rebaser.compose([]),
            revision: "root",
        }, changeFamily, () => idCompressor.generateCompressedId());
    const events = args?.events ?? (0, client_utils_1.createEmitter)();
    return new TreeCheckout(branch, false, changeFamily, schema, forest, events, mintRevisionTag, revisionTagCodec, idCompressor, args?.removedRoots, args?.logger, args?.breaker);
}
exports.createTreeCheckout = createTreeCheckout;
/**
 * An implementation of {@link ITreeCheckoutFork}.
 */
class TreeCheckout {
    constructor(_branch, 
    /** True if and only if this checkout is for a forked branch and not the "main branch" of the tree. */
    isBranch, changeFamily, storedSchema, forest, events, mintRevisionTag, revisionTagCodec, idCompressor, removedRoots = (0, index_js_2.makeDetachedFieldIndex)("repair", revisionTagCodec, idCompressor), 
    /** Optional logger for telemetry. */
    logger, breaker = new index_js_5.Breakable("TreeCheckout")) {
        this._branch = _branch;
        this.isBranch = isBranch;
        this.changeFamily = changeFamily;
        this.storedSchema = storedSchema;
        this.forest = forest;
        this.events = events;
        this.mintRevisionTag = mintRevisionTag;
        this.revisionTagCodec = revisionTagCodec;
        this.idCompressor = idCompressor;
        this.removedRoots = removedRoots;
        this.logger = logger;
        this.breaker = breaker;
        this.disposed = false;
        this.views = new Set();
        /**
         * Set of revertibles maintained for automatic disposal
         */
        this.revertibles = new Set();
        /**
         * Each branch's head commit corresponds to a revertible commit.
         * Maintaining a whole branch ensures the commit graph is not pruned in a way that would prevent the commit from
         * being reverted.
         */
        this.revertibleCommitBranches = new Map();
        /**
         * The {@link Transactor} for this checkout.
         * @remarks In the context of a checkout, transactions allow edits to be batched into atomic units.
         * Edits made during a transaction will update the local state of the tree immediately, but will be squashed into a single edit when the transaction is committed.
         * If the transaction is aborted, the local state will be reset to what it was before the transaction began.
         * Transactions may nest, meaning that a transaction may be started while a transaction is already ongoing.
         *
         * To avoid updating observers of the view state with intermediate results during a transaction,
         * use {@link ITreeCheckout#branch} and {@link ISharedTreeFork#merge}.
         */
        _TreeCheckout_transaction.set(this, new index_js_4.TransactionStack(() => {
            // Keep track of the commit that each transaction was on when it started
            // TODO:#8603: This may need to be computed differently if we allow rebasing during a transaction.
            const startCommit = this._branch.getHead();
            // Keep track of all the forks created during the transaction so that we can dispose them when the transaction ends.
            // This is a policy decision that we think is useful for the user, but it is not necessary for correctness.
            const forks = new Set();
            const onDisposeUnSubscribes = [];
            const onForkUnSubscribe = (0, index_js_4.onForkTransitive)(this, (fork) => {
                forks.add(fork);
                onDisposeUnSubscribes.push(fork.events.on("dispose", () => forks.delete(fork)));
            });
            // When each transaction is started, take a snapshot of the current state of removed roots
            const removedRoots = this.removedRoots.clone();
            this._branch.editor.enterTransaction();
            return (result) => {
                this._branch.editor.exitTransaction();
                switch (result) {
                    case index_js_4.TransactionResult.Abort:
                        this._branch.removeAfter(startCommit);
                        // If a transaction is rolled back, revert removed roots back to the latest snapshot
                        this.removedRoots = removedRoots;
                        break;
                    case index_js_4.TransactionResult.Commit: {
                        const removedCommits = this._branch.squashAfter(startCommit);
                        const transactionRevision = this._branch.getHead().revision;
                        for (const transactionStep of removedCommits) {
                            this.removedRoots.updateMajor(transactionStep.revision, transactionRevision);
                        }
                        break;
                    }
                    default:
                        (0, internal_1.unreachableCase)(result);
                }
                forks.forEach((fork) => fork.dispose());
                onDisposeUnSubscribes.forEach((unsubscribe) => unsubscribe());
                onForkUnSubscribe();
            };
        }));
        // We subscribe to `beforeChange` rather than `afterChange` here because it's possible that the change is invalid WRT our forest.
        // For example, a bug in the editor might produce a malformed change object and thus applying the change to the forest will throw an error.
        // In such a case we will crash here, preventing the change from being added to the commit graph, and preventing `afterChange` from firing.
        // One important consequence of this is that we will not submit the op containing the invalid change, since op submissions happens in response to `afterChange`.
        _branch.events.on("beforeChange", (event) => {
            if (event.change !== undefined) {
                let revision;
                if (event.type === "replace") {
                    (0, internal_1.assert)((0, index_js_5.hasSome)(event.newCommits), 0xa81 /* Expected new commit for non no-op change event */);
                    revision = (0, index_js_5.getLast)(event.newCommits).revision;
                }
                else {
                    revision = event.change.revision;
                }
                // Conflicts due to schema will be empty and thus are not applied.
                for (const change of event.change.change.changes) {
                    if (change.type === "data") {
                        const delta = (0, index_js_3.intoDelta)((0, index_js_2.tagChange)(change.innerChange, revision));
                        this.withCombinedVisitor((visitor) => {
                            (0, index_js_2.visitDelta)(delta, visitor, this.removedRoots, revision);
                        });
                    }
                    else if (change.type === "schema") {
                        // Schema changes from a current to a new schema are expected to be backwards compatible.
                        // This guarantees that all data in the forest (which is valid before the schema change)
                        // is also valid under the new schema.
                        // Note however, that such schema changes may in some cases be rolled back:
                        // Case 1: A transaction with a schema change may be aborted.
                        // The transaction may have made some data changes that would render some trees invalid
                        // under the old schema, but these changes will also be rolled back, thereby putting the forest
                        // back in the state before the transaction, which is valid under the original (reinstated) schema.
                        // Case 2: A branch with a schema change may be rebased such that the schema change (because
                        // of a constraint) is no longer applied.
                        // Such a branch may contain data changes that would render some trees invalid under the
                        // original schema. These data changes may not necessarily be rolled back.
                        // They will however be rebased over the rollback of the schema change. This rebasing will
                        // ensure that these data changes are muted if they would render some trees invalid under the
                        // original (reinstated) schema.
                        storedSchema.apply(change.innerChange.schema.new);
                    }
                    else {
                        (0, index_js_5.fail)("Unknown Shared Tree change type.");
                    }
                }
                this.events.emit("afterBatch");
            }
        });
        _branch.events.on("afterChange", (event) => {
            // The following logic allows revertibles to be generated for the change.
            // Currently only appends (including merges) and transaction commits are supported.
            if (!this.transaction.isInProgress()) {
                if (event.type === "append" ||
                    (event.type === "replace" && (0, index_js_4.getChangeReplaceType)(event) === "transactionCommit")) {
                    // TODO:#20949: When the SharedTree is detached, these commits will already have been garbage collected.
                    //       Figure out a way to generate revertibles before the commits are garbage collected.
                    for (const commit of event.newCommits) {
                        const kind = event.type === "append" ? event.kind : index_js_2.CommitKind.Default;
                        const { change, revision } = commit;
                        const getRevertible = (0, sharedTreeChangeFamily_js_1.hasSchemaChange)(change)
                            ? undefined
                            : (onRevertibleDisposed) => {
                                if (!withinEventContext) {
                                    throw new internal_2.UsageError("Cannot get a revertible outside of the context of a changed event.");
                                }
                                if (this.revertibleCommitBranches.get(revision) !== undefined) {
                                    throw new internal_2.UsageError("Cannot generate the same revertible more than once. Note that this can happen when multiple changed event listeners are registered.");
                                }
                                const revertible = this.createRevertible(revision, kind, this, onRevertibleDisposed);
                                this.revertibleCommitBranches.set(revision, _branch.fork(commit));
                                this.revertibles.add(revertible);
                                return revertible;
                            };
                        let withinEventContext = true;
                        this.events.emit("changed", { isLocal: true, kind }, getRevertible);
                        withinEventContext = false;
                    }
                }
                else if (this.isRemoteChangeEvent(event)) {
                    // TODO: figure out how to plumb through commit kind info for remote changes
                    this.events.emit("changed", { isLocal: false, kind: index_js_2.CommitKind.Default });
                }
            }
        });
        // When the branch is trimmed, we can garbage collect any repair data whose latest relevant revision is one of the
        // trimmed revisions.
        _branch.events.on("ancestryTrimmed", (revisions) => {
            this.withCombinedVisitor((visitor) => {
                revisions.forEach((revision) => {
                    // get all the roots last created or used by the revision
                    const roots = this.removedRoots.getRootsLastTouchedByRevision(revision);
                    // get the detached field for the root and delete it from the removed roots
                    for (const root of roots) {
                        visitor.destroy(this.removedRoots.toFieldKey(root), 1);
                    }
                    this.removedRoots.deleteRootsLastTouchedByRevision(revision);
                });
            });
        });
    }
    withCombinedVisitor(fn) {
        const anchorVisitor = this.forest.anchors.acquireVisitor();
        const combinedVisitor = (0, index_js_2.combineVisitors)([this.forest.acquireVisitor(), anchorVisitor], [anchorVisitor]);
        fn(combinedVisitor);
        combinedVisitor.free();
    }
    checkNotDisposed(usageError) {
        if (this.disposed) {
            if (usageError !== undefined) {
                throw new internal_2.UsageError(usageError);
            }
            (0, internal_1.assert)(false, 0x911 /* Invalid operation on a disposed TreeCheckout */);
        }
    }
    /**
     * Creates a {@link RevertibleAlpha} object that can undo a specific change in the tree's history.
     * Revision must exist in the given {@link TreeCheckout}'s branch.
     *
     * @param revision - The revision tag identifying the change to be made revertible.
     * @param kind - The {@link CommitKind} that produced this revertible (e.g., Default, Undo, Redo).
     * @param checkout - The {@link TreeCheckout} instance this revertible belongs to.
     * @param onRevertibleDisposed - Callback function that will be called when the revertible is disposed.
     * @returns - {@link RevertibleAlpha}
     */
    createRevertible(revision, kind, checkout, onRevertibleDisposed) {
        const commitBranches = checkout.revertibleCommitBranches;
        const revertible = {
            get status() {
                const revertibleCommit = commitBranches.get(revision);
                return revertibleCommit === undefined
                    ? index_js_2.RevertibleStatus.Disposed
                    : index_js_2.RevertibleStatus.Valid;
            },
            revert: (release = true) => {
                if (revertible.status === index_js_2.RevertibleStatus.Disposed) {
                    throw new internal_2.UsageError("Unable to revert a revertible that has been disposed.");
                }
                const revertMetrics = checkout.revertRevertible(revision, kind);
                checkout.logger?.sendTelemetryEvent({
                    eventName: TreeCheckout.revertTelemetryEventName,
                    ...revertMetrics,
                });
                if (release) {
                    revertible.dispose();
                }
            },
            clone: (forkedBranch) => {
                if (forkedBranch === undefined) {
                    return this.createRevertible(revision, kind, checkout, onRevertibleDisposed);
                }
                // TODO:#23442: When a revertible is cloned for a forked branch, optimize to create a fork of a revertible branch once per revision NOT once per revision per checkout.
                const forkedCheckout = (0, schematizingTreeView_js_1.getCheckout)(forkedBranch);
                const revertibleBranch = this.revertibleCommitBranches.get(revision);
                (0, internal_1.assert)(revertibleBranch !== undefined, 0xa82 /* change to revert does not exist on the given forked branch */);
                forkedCheckout.revertibleCommitBranches.set(revision, revertibleBranch.fork());
                return this.createRevertible(revision, kind, forkedCheckout, onRevertibleDisposed);
            },
            dispose: () => {
                if (revertible.status === index_js_2.RevertibleStatus.Disposed) {
                    throw new internal_2.UsageError("Unable to dispose a revertible that has already been disposed.");
                }
                checkout.disposeRevertible(revertible, revision);
                onRevertibleDisposed?.(revertible);
            },
        };
        return revertible;
    }
    viewWith(config) {
        const view = new schematizingTreeView_js_1.SchematizingSimpleTreeView(this, config, (0, index_js_3.createNodeKeyManager)(this.idCompressor), this.breaker, () => {
            this.views.delete(view);
        });
        this.views.add(view);
        return view;
    }
    get rootEvents() {
        return this.forest.anchors.events;
    }
    get editor() {
        this.checkNotDisposed();
        return this._branch.editor;
    }
    locate(anchor) {
        this.checkNotDisposed();
        return this.forest.anchors.locate(anchor);
    }
    get transaction() {
        return __classPrivateFieldGet(this, _TreeCheckout_transaction, "f");
    }
    branch() {
        this.checkNotDisposed("The parent branch has already been disposed and can no longer create new branches.");
        const anchors = new index_js_2.AnchorSet();
        const branch = this._branch.fork();
        const storedSchema = this.storedSchema.clone();
        const forest = this.forest.clone(storedSchema, anchors);
        const checkout = new TreeCheckout(branch, true, this.changeFamily, storedSchema, forest, (0, client_utils_1.createEmitter)(), this.mintRevisionTag, this.revisionTagCodec, this.idCompressor, this.removedRoots.clone(), this.logger, this.breaker);
        this.events.emit("fork", checkout);
        return checkout;
    }
    rebase(checkout) {
        this.checkNotDisposed("The target of the branch rebase has been disposed and cannot be rebased.");
        checkout.checkNotDisposed("The source of the branch rebase has been disposed and cannot be rebased.");
        (0, internal_1.assert)(!checkout.transaction.isInProgress(), 0x9af /* A view cannot be rebased while it has a pending transaction */);
        (0, internal_1.assert)(checkout.isBranch, 0xa5d /* The main branch cannot be rebased onto another branch. */);
        checkout._branch.rebaseOnto(this._branch);
    }
    rebaseOnto(checkout) {
        this.checkNotDisposed("The target of the branch rebase has been disposed and cannot be rebased.");
        checkout.rebase(this);
    }
    merge(checkout, disposeMerged = true) {
        this.checkNotDisposed("The target of the branch merge has been disposed and cannot be merged.");
        checkout.checkNotDisposed("The source of the branch merge has been disposed and cannot be merged.");
        (0, internal_1.assert)(!this.transaction.isInProgress(), 0x9b0 /* Views cannot be merged into a view while it has a pending transaction */);
        while (checkout.transaction.isInProgress()) {
            checkout.transaction.commit();
        }
        this._branch.merge(checkout._branch);
        if (disposeMerged && checkout.isBranch) {
            // Dispose the merged checkout unless it is the main branch.
            checkout[index_js_5.disposeSymbol]();
        }
    }
    updateSchema(newSchema) {
        this.checkNotDisposed();
        this.editor.schema.setStoredSchema(this.storedSchema.clone(), newSchema);
    }
    dispose() {
        this[index_js_5.disposeSymbol]();
    }
    [(_TreeCheckout_transaction = new WeakMap(), index_js_5.disposeSymbol)]() {
        this.checkNotDisposed("The branch has already been disposed and cannot be disposed again.");
        this.disposed = true;
        __classPrivateFieldGet(this, _TreeCheckout_transaction, "f").dispose();
        this.purgeRevertibles();
        this._branch.dispose();
        for (const view of this.views) {
            view.dispose();
        }
        this.events.emit("dispose");
    }
    getRemovedRoots() {
        this.assertNoUntrackedRoots();
        const trees = [];
        const cursor = this.forest.allocateCursor("getRemovedRoots");
        for (const { id, root } of this.removedRoots.entries()) {
            const parentField = this.removedRoots.toFieldKey(root);
            this.forest.moveCursorToPath({ parent: undefined, parentField, parentIndex: 0 }, cursor);
            const tree = (0, index_js_3.jsonableTreeFromCursor)(cursor);
            // This method is used for tree consistency comparison.
            const { major, minor } = id;
            const finalizedMajor = major !== undefined ? this.revisionTagCodec.encode(major) : major;
            trees.push([finalizedMajor, minor, tree]);
        }
        cursor.free();
        return trees;
    }
    /**
     * This sets the tip revision as the latest relevant revision for any removed roots that are loaded from a summary.
     * This needs to be called right after loading {@link this.removedRoots} from a summary to allow loaded data to be garbage collected.
     */
    setTipRevisionForLoadedData(revision) {
        this.removedRoots.setRevisionsForLoadedData(revision);
    }
    purgeRevertibles() {
        for (const revertible of this.revertibles) {
            revertible.dispose();
        }
    }
    disposeRevertible(revertible, revision) {
        this.revertibleCommitBranches.get(revision)?.dispose();
        this.revertibleCommitBranches.delete(revision);
        this.revertibles.delete(revertible);
    }
    revertRevertible(revision, kind) {
        if (this.transaction.isInProgress()) {
            throw new internal_2.UsageError("Undo is not yet supported during transactions.");
        }
        const revertibleBranch = this.revertibleCommitBranches.get(revision);
        (0, internal_1.assert)(revertibleBranch !== undefined, 0x7cc /* expected to find a revertible commit */);
        const commitToRevert = revertibleBranch.getHead();
        const revisionForInvert = this.mintRevisionTag();
        let change = (0, index_js_2.tagChange)(this.changeFamily.rebaser.invert(commitToRevert, false, revisionForInvert), revisionForInvert);
        const headCommit = this._branch.getHead();
        // Rebase the inverted change onto any commits that occurred after the undoable commits.
        if (commitToRevert !== headCommit) {
            change = (0, index_js_2.tagChange)((0, index_js_2.rebaseChange)(this.changeFamily.rebaser, change, commitToRevert, headCommit, this.mintRevisionTag).change, revisionForInvert);
        }
        this._branch.apply(change, kind === index_js_2.CommitKind.Default || kind === index_js_2.CommitKind.Redo
            ? index_js_2.CommitKind.Undo
            : index_js_2.CommitKind.Redo);
        // Derive some stats about the reversion to return to the caller.
        let revertAge = 0;
        let currentCommit = headCommit;
        while (commitToRevert.revision !== currentCommit.revision) {
            revertAge++;
            const parentCommit = currentCommit.parent;
            (0, internal_1.assert)(parentCommit !== undefined, 0x9a9 /* expected to find a parent commit */);
            currentCommit = parentCommit;
        }
        return { age: revertAge };
    }
    assertNoUntrackedRoots() {
        const cursor = this.forest.getCursorAboveDetachedFields();
        const rootFields = new Set([index_js_2.rootFieldKey]);
        for (const { root } of this.removedRoots.entries()) {
            rootFields.add(this.removedRoots.toFieldKey(root));
        }
        if (!cursor.firstField()) {
            return;
        }
        do {
            const field = cursor.getFieldKey();
            (0, internal_1.assert)(rootFields.has(field), 0xa22 /* Forest has a root field which is unknown to the detached field index */);
            rootFields.delete(field);
        } while (cursor.nextField());
    }
    /**
     * `true` iff the given branch change event is due to a remote change
     */
    isRemoteChangeEvent(event) {
        return (
        // Remote changes are only ever applied to the main branch
        !this.isBranch &&
            // Remote changes are applied to the main branch by rebasing it onto the trunk.
            // No other rebases are allowed on the main branch, so we can use this to detect remote changes.
            event.type === "replace" &&
            (0, index_js_4.getChangeReplaceType)(event) === "rebase");
    }
}
exports.TreeCheckout = TreeCheckout;
/**
 * The name of the telemetry event logged for calls to {@link TreeCheckout.revertRevertible}.
 * @privateRemarks Exposed for testing purposes.
 */
TreeCheckout.revertTelemetryEventName = "RevertRevertible";
//# sourceMappingURL=treeCheckout.js.map